@@ -9,7 +9,7 @@ use std::collections::HashMap;
 use std::fs::File;
 use std::io::BufReader;
 use std::path::Path;
-use std::thread;
+use std::thread::{self, JoinHandle};
 use std::time::SystemTime;
 static INTERNAL_SAMPLE_RATE: usize = 48000;
 pub type VadMode = webrtc_vad::VadMode;
@@ -19,7 +19,10 @@ pub struct WakewordDetectorBuilder {
     sample_format: Option<SampleFormat>,
     bits_per_sample: Option<u16>,
     vad_mode: Option<VadMode>,
-    vad_detection_delay: Option<u16>,
+    vad_sensitivity: Option<f32>,
+    vad_delay: Option<u16>,
+    eager_mode: bool,
+    single_thread: bool,
     threshold: Option<f32>,
     comparator_band_size: Option<usize>,
     comparator_ref: Option<f32>,
@@ -32,8 +35,11 @@ impl WakewordDetectorBuilder {
             sample_format: None,
             bits_per_sample: None,
             // detection options
+            eager_mode: false,
+            single_thread: false,
             vad_mode: None,
-            vad_detection_delay: None,
+            vad_delay: None,
+            vad_sensitivity: None,
             threshold: None,
             comparator_band_size: None,
             comparator_ref: None,
@@ -44,14 +50,18 @@ impl WakewordDetectorBuilder {
             self.get_sample_rate(),
             self.get_sample_format(),
             self.get_bits_per_sample(),
+            self.get_eager_mode(),
+            self.get_single_thread(),
             self.get_vad_mode(),
-            self.get_vad_detection_delay(),
+            self.get_vad_delay(),
+            self.get_vad_sensitivity(),
             self.get_threshold(),
             self.get_comparator_band_size(),
             self.get_comparator_ref(),
         )
     }
     pub fn set_threshold(&mut self, value: f32) {
+        assert!(value >= 0. || value <= 1.);
         self.threshold = Some(value);
     }
     pub fn set_bits_per_sample(&mut self, value: u16) {
@@ -69,8 +79,18 @@ impl WakewordDetectorBuilder {
     pub fn set_comparator_ref(&mut self, value: f32) {
         self.comparator_ref = Some(value);
     }
-    pub fn set_vad_detection_delay(&mut self, value: u16) {
-        self.vad_detection_delay = Some(value);
+    pub fn set_eager_mode(&mut self, value: bool) {
+        self.eager_mode = value;
+    }
+    pub fn set_single_thread(&mut self, value: bool) {
+        self.single_thread = value;
+    }
+    pub fn set_vad_delay(&mut self, value: u16) {
+        self.vad_delay = Some(value);
+    }
+    pub fn set_vad_sensitivity(&mut self, value: f32) {
+        assert!(value >= 0. || value <= 1.);
+        self.vad_sensitivity = Some(value);
     }
     pub fn set_vad_mode(&mut self, value: VadMode) {
         self.vad_mode = Some(value);
@@ -93,8 +113,17 @@ impl WakewordDetectorBuilder {
     fn get_comparator_ref(&self) -> f32 {
         self.comparator_ref.unwrap_or(0.22)
     }
-    fn get_vad_detection_delay(&self) -> u16 {
-        self.vad_detection_delay.unwrap_or(3)
+    fn get_eager_mode(&self) -> bool {
+        self.eager_mode
+    }
+    fn get_single_thread(&self) -> bool {
+        self.single_thread
+    }
+    fn get_vad_sensitivity(&self) -> f32 {
+        self.vad_sensitivity.unwrap_or(0.5)
+    }
+    fn get_vad_delay(&self) -> u16 {
+        self.vad_delay.unwrap_or(3)
     }
     fn get_vad_mode(&self) -> Option<VadMode> {
         if self.vad_mode.is_none() {
@@ -110,16 +139,18 @@ impl WakewordDetectorBuilder {
 }
 pub struct WakewordDetector {
     // input options
-    sample_rate: usize,
     sample_format: SampleFormat,
     bits_per_sample: u16,
     // detection options
     threshold: f32,
     comparator_band_size: usize,
     comparator_ref: f32,
+    eager_mode: bool,
+    single_thread: bool,
     resampler: Option<FftFixedInOut<f32>>,
     vad_detector: Option<webrtc_vad::Vad>,
-    vad_detection_delay: u16,
+    vad_delay: u16,
+    vad_sensitivity: f32,
     // state
     samples_per_frame: usize,
     buffering: bool,
@@ -141,8 +172,11 @@ impl WakewordDetector {
         sample_format: SampleFormat,
         bits_per_sample: u16,
         // detection options
+        eager_mode: bool,
+        single_thread: bool,
         vad_mode: Option<VadMode>,
-        vad_detection_delay: u16,
+        vad_delay: u16,
+        vad_sensitivity: f32,
         threshold: f32,
         comparator_band_size: usize,
         comparator_ref: f32,
@@ -167,7 +201,6 @@ impl WakewordDetector {
         };
         let detector = WakewordDetector {
             threshold,
-            sample_rate,
             sample_format,
             bits_per_sample,
             samples_per_frame,
@@ -186,10 +219,13 @@ impl WakewordDetector {
                 None
             },
             resampler,
+            eager_mode,
+            single_thread,
             voice_detections: Vec::with_capacity(100),
             audio_cache: Vec::with_capacity(100),
             vad_detector,
-            vad_detection_delay,
+            vad_delay,
+            vad_sensitivity,
             voice_detection_time: SystemTime::UNIX_EPOCH,
         };
         detector
@@ -371,8 +407,7 @@ impl WakewordDetector {
     }
     fn apply_vad_detection(&mut self, resampled_audio: Vec<f32>) -> Option<DetectedWakeword> {
         if self.vad_detector.is_some()
-            && self.voice_detection_time.elapsed().unwrap().as_secs()
-                >= self.vad_detection_delay as u64
+            && self.voice_detection_time.elapsed().unwrap().as_secs() >= self.vad_delay as u64
         {
             let vad = self.vad_detector.as_mut().unwrap();
             let is_voice_result = vad.is_voice_segment(
@@ -386,35 +421,39 @@ impl WakewordDetector {
             }
             self.voice_detections
                 .push(is_voice_result.is_err() || is_voice_result.unwrap());
-            let voice_activity = self.voice_detections.iter().filter(|i| **i == true).count();
-            if voice_activity < 50 {
+            if self.voice_detections.iter().filter(|i| **i == true).count()
+                >= (self.vad_sensitivity * 100.) as usize
+            {
+                debug!("voice detected; processing cache");
+                self.audio_cache
+                    .drain(0..self.audio_cache.len())
+                    .collect::<Vec<_>>()
+                    .into_iter()
+                    .for_each(|i| {
+                        self.process_encoded_audio(&i, false);
+                    });
+                self.voice_detections.clear();
+                self.voice_detection_time = SystemTime::now();
+                debug!("detection time updated, processing last frame");
+                self.process_encoded_audio(&resampled_audio, true)
+            } else {
                 if self.audio_cache.len() >= self.min_frames {
                     self.audio_cache
                         .drain(0..=self.audio_cache.len() - self.min_frames);
                 }
                 self.audio_cache.push(resampled_audio);
                 None
-            } else {
-                debug!("voice detected; processing cache");
-                self.audio_cache.to_vec().into_iter().for_each(|i| {
-                    self.process_encoded_audio(i, false);
-                });
-                self.voice_detections.clear();
-                self.audio_cache.clear();
-                self.voice_detection_time = SystemTime::now();
-                debug!("detection time updated, processing last frame");
-                self.process_encoded_audio(resampled_audio, true)
             }
         } else {
-            self.process_encoded_audio(resampled_audio, true)
+            self.process_encoded_audio(&resampled_audio, true)
         }
     }
     fn process_encoded_audio(
         &mut self,
-        resampled_audio: Vec<f32>,
+        audio_chunk: &[f32],
         run_detection: bool,
     ) -> Option<DetectedWakeword> {
-        self.extractor.shift_and_filter_input(&resampled_audio);
+        self.extractor.shift_and_filter_input(&audio_chunk);
         self.extractor.compute_frame_features();
         let features = self.extractor.features().to_vec();
         let mut detection: Option<DetectedWakeword> = None;
@@ -467,7 +506,7 @@ impl WakewordDetector {
                     })
                     .collect::<Vec<_>>();
                 if sample_rate as usize != INTERNAL_SAMPLE_RATE {
-                    self.resample_audio(&samples)
+                    resample_audio(sample_rate as usize, &samples)
                 } else {
                     samples
                 }
@@ -478,7 +517,7 @@ impl WakewordDetector {
                     .map(|s| s.map(|s| s * 32767.0).unwrap())
                     .collect::<Vec<_>>();
                 if sample_rate as usize != INTERNAL_SAMPLE_RATE {
-                    self.resample_audio(&samples)
+                    resample_audio(sample_rate as usize, &samples)
                 } else {
                     samples
                 }
@@ -503,44 +542,46 @@ impl WakewordDetector {
             .collect::<Vec<_>>();
         Ok(self.normalize_features(all_features))
     }
-    fn resample_audio(&self, audio_pcm_signed: &[f32]) -> Vec<f32> {
-        let mut resampler =
-            FftFixedInOut::<f32>::new(self.sample_rate, INTERNAL_SAMPLE_RATE, 480, 1).unwrap();
-        let mut out = resampler.output_buffer_allocate();
-        let resampled_audio = audio_pcm_signed
-            .chunks_exact(resampler.input_frames_next())
-            .map(|sample| {
-                resampler
-                    .process_into_buffer(&[sample], &mut out[..], None)
-                    .unwrap();
-                out.get(0).unwrap().to_vec()
-            })
-            .flatten()
-            .collect::<Vec<f32>>();
-        resampled_audio
-    }
     fn run_detection(&mut self) -> Option<DetectedWakeword> {
         let features = self.normalize_features(self.frames.to_vec());
         let result_option = self.get_best_keyword(features);
         match result_option {
             Some(result) => {
-                if self.result_state.is_some() {
-                    let prev_wakeword = self.result_state.as_ref().unwrap().wakeword.clone();
-                    let prev_score = self.result_state.as_ref().unwrap().score;
-                    if result.wakeword == prev_wakeword && result.score < prev_score {
-                        debug!(
-                            "keyword '{}' detected, score {}",
-                            result.wakeword, prev_score
-                        );
-                        self.reset();
-                        return Some(DetectedWakeword {
-                            wakeword: result.wakeword,
-                            score: prev_score,
-                        });
+                if self.eager_mode {
+                    if result.index != 0 {
+                        debug!("Sorting '{}' templates", result.wakeword);
+                        self.keywords
+                            .get_mut(&result.wakeword)
+                            .unwrap()
+                            .prioritize_template(result.index);
                     }
+                    debug!(
+                        "keyword '{}' detected, score {}",
+                        result.wakeword, result.score
+                    );
+                    self.reset();
+                    Some(result)
+                } else {
+                    if self.result_state.is_some() {
+                        let prev_wakeword = self.result_state.as_ref().unwrap().wakeword.clone();
+                        let prev_score = self.result_state.as_ref().unwrap().score;
+                        let prev_index = self.result_state.as_ref().unwrap().index;
+                        if result.wakeword == prev_wakeword && result.score < prev_score {
+                            debug!(
+                                "keyword '{}' detected, score {}",
+                                result.wakeword, prev_score
+                            );
+                            self.reset();
+                            return Some(DetectedWakeword {
+                                wakeword: result.wakeword,
+                                score: prev_score,
+                                index: prev_index,
+                            });
+                        }
+                    }
+                    self.result_state = Some(result);
+                    None
                 }
-                self.result_state = Some(result);
-                None
             }
             None => None,
         }
@@ -579,52 +620,55 @@ impl WakewordDetector {
         normalized_frames
     }
     fn get_best_keyword(&self, features: Vec<Vec<f32>>) -> Option<DetectedWakeword> {
-        let mut results: Vec<thread::JoinHandle<Option<DetectedWakeword>>> = vec![];
-        for (name, keyword) in &self.keywords {
-            if !keyword.is_enabled() {
-                continue;
-            }
-            let wakeword_name = name.clone();
-            let threshold = keyword.get_threshold().unwrap_or(self.threshold);
-            let templates = keyword.get_templates();
-            let features_copy = features.to_vec();
-            let comparator_band_size = self.comparator_band_size;
-            let comparator_ref = self.comparator_ref;
-            let comparator_task = thread::spawn(move || {
-                let comparator = FeatureComparator::new(comparator_band_size, comparator_ref);
-                let mut detection: Option<DetectedWakeword> = None;
-                for template in templates {
-                    let mut frames = features_copy.to_vec();
-                    if frames.len() > template.len() {
-                        frames.drain(template.len()..frames.len());
-                    }
-                    let score = comparator.compare(
-                        &template.iter().map(|item| &item[..]).collect::<Vec<_>>(),
-                        &frames.iter().map(|item| &item[..]).collect::<Vec<_>>(),
-                    );
-                    if score < threshold {
-                        continue;
-                    }
-                    let prev_detection = detection.as_ref();
-                    if prev_detection.is_some() && score < prev_detection.unwrap().score {
-                        continue;
-                    }
-                    detection = Some(DetectedWakeword {
-                        wakeword: wakeword_name.clone(),
-                        score,
-                    });
-                }
-                detection
-            });
-            results.push(comparator_task);
-        }
-        let mut detections: Vec<DetectedWakeword> = vec![];
-        for result_task in results {
-            let result = result_task.join().unwrap();
-            if result.is_some() {
-                detections.push(result.unwrap());
-            }
-        }
+        let mut detections: Vec<DetectedWakeword> =
+            if self.single_thread || self.keywords.len() <= 1 {
+                self.keywords
+                    .iter()
+                    .filter_map(|(name, keyword)| {
+                        let templates = keyword.get_templates();
+                        let threshold = keyword.get_threshold().unwrap_or(self.threshold);
+                        run_wakeword_detection(
+                            self.comparator_band_size,
+                            self.comparator_ref,
+                            &templates,
+                            self.eager_mode,
+                            &features,
+                            threshold,
+                            name.clone(),
+                        )
+                    })
+                    .collect::<Vec<_>>()
+            } else {
+                self.keywords
+                    .iter()
+                    .filter_map(|(name, keyword)| {
+                        if !keyword.is_enabled() {
+                            None
+                        } else {
+                            let wakeword_name = name.clone();
+                            let threshold = keyword.get_threshold().unwrap_or(self.threshold);
+                            let templates = keyword.get_templates();
+                            let comparator_band_size = self.comparator_band_size;
+                            let comparator_ref = self.comparator_ref;
+                            let eager_mode = self.eager_mode;
+                            let features_copy = features.to_vec();
+                            Some(thread::spawn(move || {
+                                run_wakeword_detection(
+                                    comparator_band_size,
+                                    comparator_ref,
+                                    &templates,
+                                    eager_mode,
+                                    &features_copy,
+                                    threshold,
+                                    wakeword_name,
+                                )
+                            }))
+                        }
+                    })
+                    .map(JoinHandle::join)
+                    .filter_map(Result::unwrap)
+                    .collect::<Vec<_>>()
+            };
         if detections.is_empty() {
             None
         } else {
@@ -636,16 +680,76 @@ impl WakewordDetector {
         self.samples_per_frame
     }
 }
+fn resample_audio(input_sample_rate: usize, audio_pcm_signed: &[f32]) -> Vec<f32> {
+    let mut resampler =
+        FftFixedInOut::<f32>::new(input_sample_rate, INTERNAL_SAMPLE_RATE, 480, 1).unwrap();
+    let mut out = resampler.output_buffer_allocate();
+    let resampled_audio = audio_pcm_signed
+        .chunks_exact(resampler.input_frames_next())
+        .map(|sample| {
+            resampler
+                .process_into_buffer(&[sample], &mut out[..], None)
+                .unwrap();
+            out.get(0).unwrap().to_vec()
+        })
+        .flatten()
+        .collect::<Vec<f32>>();
+    resampled_audio
+}
+
+fn run_wakeword_detection(
+    comparator_band_size: usize,
+    comparator_ref: f32,
+    templates: &[Vec<Vec<f32>>],
+    eager_mode: bool,
+    features: &[Vec<f32>],
+    threshold: f32,
+    wakeword_name: String,
+) -> Option<DetectedWakeword> {
+    let comparator = FeatureComparator::new(comparator_band_size, comparator_ref);
+    let mut detection: Option<DetectedWakeword> = None;
+    for (index, template) in templates.iter().enumerate() {
+        if eager_mode && detection.is_some() {
+            break;
+        }
+        let mut frames = features.to_vec();
+        if frames.len() > template.len() {
+            frames.drain(template.len()..frames.len());
+        }
+        let score = comparator.compare(
+            &template.iter().map(|item| &item[..]).collect::<Vec<_>>(),
+            &frames.iter().map(|item| &item[..]).collect::<Vec<_>>(),
+        );
+        if score < threshold {
+            continue;
+        }
+        let prev_detection = detection.as_ref();
+        if prev_detection.is_some() && score < prev_detection.unwrap().score {
+            continue;
+        }
+        detection = Some(DetectedWakeword {
+            wakeword: wakeword_name.clone(),
+            score,
+            index,
+        });
+    }
+    detection
+}
 
 pub struct DetectedWakeword {
+    // Detected wakeword name
     pub wakeword: String,
+    // Detection score
     pub score: f32,
+    // Detected wakeword template index
+    pub index: usize,
 }
 impl Clone for DetectedWakeword {
     fn clone(&self) -> Self {
         Self {
             wakeword: self.wakeword.clone(),
             score: self.score.clone(),
+            index: self.index.clone(),
         }
     }
 }
